(*
StartupAce StartupEntry.pas Unit (class TStartupEntry and related)
Copyright (C) 2006  Jakob Reschke

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this software; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*)

unit StartupEntry;

interface

uses
  paths, MessageStrs;

type
  TStartupEntry = class;
  PStartupEntry = ^TStartupEntry;

  TRunType = (rtStartupAce, rtRegistry, rtLnk, rtNone);

  TStartupEntryEvent = procedure (Sender: TStartupEntry) of object;

  TStartupEntry = class
  private
    FPath,
    FArguments,
    FCaption: WideString;
    FActive,
    FAllUsers: Boolean;
    FDelay: Cardinal;
    FRunType: TRunType;
    FOnChanged: TStartupEntryEvent;
    FWaitFor: WideString;
    OrgPath: WideString;
    OrgParameters: WideString;
    OrgCaption: WideString;
    OrgActive,
    OrgAllUsers: Boolean;
    OrgDelay: Cardinal;
    OrgRunType: TRunType;
    OrgWaitFor: WideString;
    FDeleted: Boolean;
    FTreeNode: Pointer;
    procedure SetOnChanged(const Value: TStartupEntryEvent);
    procedure SetTreeNode(const Value: Pointer);
    function IsStartupAcesEntry: Boolean;
    procedure SetWaitFor(const Value: WideString);
  protected
    function GetChanged: Boolean;
    procedure SetActive(const Value: Boolean);
    procedure SetDelay(const Value: Cardinal);
    procedure SetDeleted(const Value: Boolean);
    procedure SetAllUsers(const Value: Boolean);
    procedure SetCaption(const Value: WideString);
    procedure SetRunType(const Value: TRunType);
    procedure SetArguments(const Value: WideString);
    procedure SetPath(const Value: WideString);
    function GetEditable: Boolean;
  public
    constructor Create(TreeNode: Pointer);
    procedure ParseRegString(const RegString: WideString);
    procedure SetCurrentAsOriginalValues;
    procedure SetOrgValues(Path, Params, Caption: WideString; Active: Boolean; RunType: TRunType; Delay: Cardinal; AllUsers: Boolean; WaitFor: WideString);
    procedure Reset;
    function Save(DeleteOldEntry, CreateNewEntry: Boolean): Boolean;
    property Path: WideString read FPath write SetPath;
    property Arguments: WideString read FArguments write SetArguments;
    property Caption: WideString  read FCaption write SetCaption;
    property Active: Boolean read FActive write SetActive;    
    property AllUsers: Boolean read FAllUsers write SetAllUsers;
    property RunType: TRunType read FRunType write SetRunType;
    property Delay: Cardinal read FDelay write SetDelay;
    property Deleted: Boolean read FDeleted write SetDeleted;
    property Changed: Boolean read GetChanged;
    property WaitFor: WideString read FWaitFor write SetWaitFor;
    property OriginalCaption: WideString read OrgCaption;
    property OriginalAllUsers: Boolean read OrgAllUsers;
    property OriginalRunType: TRunType read OrgRunType;
    property OriginalDelay: Cardinal read OrgDelay;
    property Editable: Boolean read GetEditable;
    property OnChanged: TStartupEntryEvent read FOnChanged write SetOnChanged;
    property TreeNode: Pointer read FTreeNode write SetTreeNode;
  end;

implementation

uses
  Windows, Contnrs, SysUtils, Policy, Registry, Statics, Forms, ShlObj, ComObj, 
  ActiveX;

constructor TStartupEntry.Create(TreeNode: Pointer);
begin
  FPath := '';
  FArguments := '';
  FCaption := '';
  FDelay := 0;
  FWaitFor := '';
  FActive := true;
  FRunType := rtNone;
  FTreeNode := TreeNode;
end;

procedure TStartupEntry.SetCurrentAsOriginalValues;
begin
  SetOrgValues(FPath, FArguments, FCaption, FActive, FRunType, FDelay,
    FAllUsers, FWaitFor);
end;

function TStartupEntry.GetChanged: Boolean;
begin
  Result := not (true
    and (FRunType = OrgRunType)
    and (FAllUsers = OrgAllUsers)
    and (FActive = OrgActive)
    and (FDelay = OrgDelay)
    and (FPath = OrgPath)
    and (FArguments = OrgParameters)
    and (FCaption = OrgCaption)
    and (FWaitFor = OrgWaitFor)
    and not Deleted
    );
end;

function TStartupEntry.GetEditable: Boolean;
begin
  Result := (not IsStartupAcesEntry) and (not FAllUsers or IsAdmin); 
end;

function TStartupEntry.IsStartupAcesEntry: Boolean;
begin
  Result := (FCaption = 'StartupAce') and not FAllUsers;
end;

procedure TStartupEntry.Reset;
begin
  FCaption := OrgCaption;
  FPath := OrgPath;
  FArguments := OrgParameters;
  FAllUsers:= OrgAllUsers;
  FActive := OrgActive;
  FDelay := OrgDelay;
  FRunType := OrgRunType;
  FWaitFor := OrgWaitFor;
  FDeleted := false; 
  if Assigned(FOnChanged) then FOnChanged(self);
end;

function TStartupEntry.Save(DeleteOldEntry, CreateNewEntry: Boolean): Boolean;
var
  reg: TRegistry;
  s: WideString;
  IObj: IUnknown;
  Link: IShellLinkW;
  IPFile: IPersistFile;
  StrBuf: array of WideChar;
  Args: array [0..0] of WideString;
begin
  if not CreateNewEntry and not DeleteOldEntry then
    raise EInvalidOp.Create('Senseless function call: TStartupEntry.Save(False, False)!');
  Result := False;
  if Self.Changed then
  begin
    reg := TRegistry.Create;
    try
      {$REGION 'Delete old entry'}
      if DeleteOldEntry then
      begin
        if Self.Deleted
          or (Self.OriginalCaption <> Self.Caption)
          or (Self.OriginalAllUsers <> Self.AllUsers)
          or (Self.OriginalRunType <> Self.RunType) then
        begin
          case self.OriginalRunType of
            rtStartupAce:
            begin
              if OriginalAllUsers then
                reg.RootKey := HKEY_LOCAL_MACHINE
              else
                reg.RootKey := HKEY_CURRENT_USER;
              if reg.OpenKey(RegistryKeys.StartupAce, false) then
              begin
                reg.DeleteKey(OriginalCaption);
                reg.CloseKey;
              end
              else
              begin
                MessageBoxW(Application.Handle,
                  PWideChar(MessageStrings.CannotAccessStartupAceKey),
                  PWideChar(MessageStrings.AccessDenied),
                  MB_OK or MB_ICONERROR);
              end;
            end;
            rtRegistry:
            begin
              if OriginalAllUsers then
                reg.RootKey := HKEY_LOCAL_MACHINE
              else
                reg.RootKey := HKEY_CURRENT_USER;
              if reg.OpenKey(RegistryKeys.Run, false) then
              begin
                reg.DeleteValue(OriginalCaption);
                reg.CloseKey;
              end
              else
              begin
                MessageBoxW(Application.Handle,
                  PWideChar(MessageStrings.CannotAccessRunKey),
                  PWideChar(MessageStrings.AccessDenied),
                  MB_OK or MB_ICONERROR);
              end;
            end;
            rtLnk:
            begin
              if OriginalAllUsers then
                DeleteFileW(PWideChar(
                  GetSpecialFolder(CSIDL_COMMON_STARTUP) + '\' +
                  OriginalCaption + '.lnk'))
              else
                DeleteFileW(PWideChar(
                  GetSpecialFolder(CSIDL_STARTUP) + '\' +
                  OriginalCaption + '.lnk'));
            end;
          end;
        end;
        if not CreateNewEntry then Result := True;
      end;
      {$ENDREGION}
      {$REGION 'Write new entry'}
      if CreateNewEntry then
      begin
        if not Self.Deleted then
        begin
          case Self.RunType of
            rtStartupAce:
            begin
              if AllUsers then
                reg.RootKey := HKEY_LOCAL_MACHINE
              else
                reg.RootKey := HKEY_CURRENT_USER;
              if reg.OpenKey(RegistryKeys.StartupAce + '\' + Caption, true) then
              begin
                reg.WriteString('Path', UTF8Encode(Path));
                reg.WriteString('Parameters', UTF8Encode(Arguments));
                reg.WriteInteger('Delay', Delay);
                reg.WriteString('Active', BoolToStr(Active));
                if self.WaitFor = '' then
                begin
                  if reg.ValueExists('WaitFor') then
                    reg.DeleteValue('WaitFor');
                end
                else
                begin
                  reg.WriteString('WaitFor', UTF8Encode(self.WaitFor));
                end;
                reg.CloseKey;
              end
              else
              begin
                Args[0] := Self.Caption;
                SetLength(StrBuf,
                  Length(MessageStrings.CannotAccessEntrysRegistryKey) - 2  // '%1'
                  + Length(Self.Caption));
                FormatMessageW(
                  FORMAT_MESSAGE_FROM_STRING or FORMAT_MESSAGE_ARGUMENT_ARRAY,
                  PWideChar(MessageStrings.CannotAccessEntrysRegistryKey),
                  0, 0, @StrBuf[0], Length(StrBuf), @Args[0]);
                MessageBoxW(Application.Handle,
                  PWideChar(StrBuf),
                  PWideChar(MessageStrings.AccessDenied),
                  MB_OK or MB_ICONERROR);
              end;
            end;
            rtRegistry:
            begin
              if AllUsers then
                reg.RootKey := HKEY_LOCAL_MACHINE
              else
                reg.RootKey := HKEY_CURRENT_USER;
              if reg.OpenKey(RegistryKeys.Run, true) then
              begin
                s := '"' + Path + '"';
                if Length(Trim(Arguments)) > 0 then
                  s := s + ' ' + Arguments;
                if not Active then
                  s := 'REM ' + s;
                reg.WriteString(Caption, s);
                s := '';
                reg.CloseKey;
              end
              else
              begin
                MessageBoxW(Application.Handle,
                  PWideChar(MessageStrings.CannotAccessRunKey),
                  nil,
                  MB_OK or MB_ICONERROR);
              end;
            end;
            rtLnk:
            begin
              IObj := CreateComObject(CLSID_ShellLink);
              Link := IObj as IShellLinkW;
              IPFile := IObj as IPersistFile;
              with Link do
              begin
                if Active then
                begin
                  SetPath(PWideChar(Path));
                  SetArguments(PWideChar(Arguments));
                end
                else
                begin
                  SetPath('REM');
                  SetArguments(PWideChar('"' + Path + '" ' + Arguments));
                end;
              end;
              if AllUsers then
                IPFile.Save(PWideChar(GetSpecialFolder(CSIDL_COMMON_STARTUP) + '\' +
                  Caption + '.lnk'), false)
              else
                IPFile.Save(PWideChar(GetSpecialFolder(CSIDL_STARTUP) + '\' +
                  Caption + '.lnk'), false);
            end;
          end;
          reg.CloseKey;
          Self.SetCurrentAsOriginalValues;
        end;
        Result := True;
      end;
      {$ENDREGION}
    finally
      reg.Free;
    end;
  end
  else
  begin
    Result := True;
  end;
end;

procedure TStartupEntry.SetActive(const Value: Boolean);
begin
  FActive := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetCaption(const Value: WideString);
begin
  FCaption := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetDelay(const Value: Cardinal);
begin
  FDelay := Value;  
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetDeleted(const Value: Boolean);
begin
  FDeleted := Value;  
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetOnChanged(const Value: TStartupEntryEvent);
begin
  FOnChanged := Value;
end;

procedure TStartupEntry.SetOrgValues(Path, Params, Caption: WideString;
  Active: Boolean; RunType: TRunType; Delay: Cardinal; AllUsers: Boolean;
  WaitFor: WideString);
begin
  OrgPath := Path;
  OrgParameters := Params;
  OrgCaption := Caption;
  OrgActive := Active;
  OrgDelay := Delay;
  OrgRunType := RunType;
  OrgAllUsers := AllUsers;
  OrgWaitFor := WaitFor;
end;

procedure TStartupEntry.SetArguments(const Value: WideString);
begin
  FArguments := Value; 
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetPath(const Value: WideString);
begin
  FPath := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetRunType(const Value: TRunType);
begin
  FRunType := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetAllUsers(const Value: Boolean);
begin
  FAllUsers := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.SetTreeNode(const Value: Pointer);
begin
  FTreeNode := Value;
end;

procedure TStartupEntry.SetWaitFor(const Value: WideString);
begin
  FWaitFor := Value;
  if Assigned(FOnChanged) then FOnChanged(self);
end;

procedure TStartupEntry.ParseRegString(const RegString: WideString);
var
  quoteOpen, strOpen, pathDone: Boolean;
  I, startpos, endpos, l: Integer;
begin
  quoteOpen := false;
  strOpen := false;
  pathDone := false;
  startpos := -1;
  l := Length(RegString);
  Active := CompareStringW(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
    PWideChar(Copy(RegString, 1, 3)), 3, 'REM', 3) <> CSTR_EQUAL;
  for I := 1 to Length(RegString) do
  begin
    if (not Active) and (I <= 3) then
      continue;

    if not pathDone then
    begin
      if RegString[I] = '"' then
      begin
        if quoteOpen then  // pfad fing bereits wo anders an
        begin
          // der Pfad endet
          quoteOpen := false;
          strOpen := false;
          endpos := I;
          Path := copy(RegString, startpos, endpos - startpos);
          pathDone := true;
        end
        else
        begin
          // der Pfad fngt an
          quoteOpen := true;
          startpos := I + 1;
        end;
      end;
      if not pathDone and not quoteOpen then
      begin
        if RegString[I] = ' ' then
        begin
          if strOpen then
          begin
            // die Zeichenkette hrt auf
            strOpen := false;
            endpos := I;
            Path := Copy(RegString, startpos, endpos - startpos);
            if not TPath.Exists(TPath.Expand(Path)) then
            begin
              // Pfad existiert nicht -> nchstes Leerzeichen abwarten
              strOpen := true;
            end else
              pathDone := true;
          end;
        end
        else if not strOpen then
        begin
          // die Zeichenkette fngt an
          strOpen := true;
          startpos := I;
        end;
      end;
    end
    else
    begin
      if not strOpen and (RegString[I] <> ' ') then
      begin
        strOpen := true;
        Arguments := copy(RegString, I, l + 1 - I);
      end;
    end;
    if (I = Length(RegString)) and not pathDone then
    begin
      Path := Copy(RegString, startpos, l + 1 - startpos);
      pathDone := true;
    end;
  end;
end;

end.
